from typing import List, Pattern
from importlib_resources import files
from re import compile
from pandas import concat, read_csv

from eis1600.helper.Singleton import Singleton
from openiti.helper.ara import denormalize

path = files('eis1600.gazetteers.data').joinpath('onomastic_gazetteer.csv')


@Singleton
class Onomastics:
    """
    Gazetteer

    :ivar __end List[str]: List of terms which indicate the end of the NASAB.
    :ivar __exp List[str]: List of explanatory terms used in the onomastic information.
    :ivar __ism List[str]: List of isms/first names.
    :ivar __laq List[str]: List of laqabs (including mamluk shortforms).
    :ivar __nsb List[str]: List of nisbas.
    :ivar __swm List[str]: List of professions.
    :ivar __tot List[str]: List of all onomastic terms: _exp + _ism + _laq + _nsb + _swm
    :ivar __rpl List[Tuple[str, str]]: List of tuples: expression and its replacement.
    :ivar __ngrams DataFrame: Subset of the df containing all name parts which are ngrams with n > 1.
    :ivar __ngrams_regex re: re of all name parts which are ngrams with n > 1, sorted from longest to shortest ngrams.
    """
    __end = None
    __exp = None
    __ism = None
    __kun = None
    __laq = None
    __nsb = None
    __swm = None
    __tot = None
    __ngrams = None
    __ngrams_regex = None

    def __init__(self) -> None:
        oa_df = read_csv(path)
        oa_df['NGRAM'] = oa_df['NGRAM'].astype('uint8')
        oa_df['CATEGORY'] = oa_df['CATEGORY'].astype('category')

        Onomastics.__end = oa_df.loc[oa_df['CATEGORY'] == 'END', 'VALUE'].to_list()
        Onomastics.__exp = oa_df.loc[oa_df['CATEGORY'] == 'EXP', 'VALUE'].to_list()
        Onomastics.__ism = oa_df.loc[oa_df['CATEGORY'] == 'ISM', 'VALUE'].to_list()
        Onomastics.__ism = oa_df.loc[oa_df['CATEGORY'] == 'KUN', 'VALUE'].to_list()
        Onomastics.__laq = oa_df.loc[oa_df['CATEGORY'] == 'LAQ', 'VALUE'].to_list()
        Onomastics.__nsb = oa_df.loc[oa_df['CATEGORY'] == 'NSB', 'VALUE'].to_list()
        Onomastics.__swm = oa_df.loc[oa_df['CATEGORY'] == 'SWM', 'VALUE'].to_list()

        Onomastics.__tot = Onomastics.__ism + Onomastics.__laq + Onomastics.__nsb + Onomastics.__swm + Onomastics.__exp
        expression = Onomastics.__exp + Onomastics.__swm + Onomastics.__ism

        def rplc_to_aba(row):
            new_row = row
            new_row['VALUE'] = row['VALUE'].replace('أبو', 'أبا')
            return new_row

        def rplc_to_abi(row):
            new_row = row
            new_row['VALUE'] = row['VALUE'].replace('أبو', 'أبي')
            return new_row

        abu_rows = oa_df.loc[oa_df['VALUE'].str.contains('أبو')]
        spelling_variations_1 = abu_rows.apply(rplc_to_aba, axis=1)
        spelling_variations_2 = abu_rows.apply(rplc_to_abi, axis=1)
        df_abu_variations = concat([oa_df.loc[oa_df['CATEGORY'] != 'END'], spelling_variations_1,
                                       spelling_variations_2])

        # Sort from longest to shortest ngrams - longest need to come first in regex otherwise only the shorter one
        # will be matched
        Onomastics.__ngrams = df_abu_variations.sort_values(by=['NGRAM'], ascending=False)
        ngrams = Onomastics.__ngrams['VALUE'].to_list()
        Onomastics.__ngrams_regex = compile('(^| )(' + denormalize('|'.join(ngrams)) + ')')

    @staticmethod
    def exp() -> List[str]:
        return Onomastics.__exp

    @staticmethod
    def end() -> List[str]:
        return Onomastics.__end

    @staticmethod
    def total() -> List[str]:
        return Onomastics.__tot

    @staticmethod
    def get_ngrams_regex() -> Pattern[str]:
        return Onomastics.__ngrams_regex

    @staticmethod
    def get_ngram_tag(ngram: str) -> str:
        lookup = Onomastics.__ngrams.loc[Onomastics.__ngrams['VALUE'].str.fullmatch(denormalize(ngram))]
        if len(lookup) > 1:
            all_pos = ['Ü' + cat + str(n) for cat, n in zip(lookup['CATEGORY'].to_list(), lookup['NGRAM'].to_list())]
            return '___'.join(all_pos) + ' '
        elif len(lookup) == 1:
            return 'Ü' + str(lookup.iloc[0]['CATEGORY']) + str(lookup.iloc[0]['NGRAM']) + ' '
        else:
            return 'ÜNaN' + str(len(ngram)) + ' '



from sys import argv, exit
from os.path import isfile, splitext
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from functools import partial

from p_tqdm import p_uimap
from tqdm import tqdm

from eis1600.helper.repo import get_path_to_other_repo, read_files_from_autoreport, get_files_from_eis1600_dir, \
    write_to_readme, TEXT_REPO
from eis1600.miu.methods import disassemble_text


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.EIS1600':
                parser.error('You need to input an EIS1600 file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to disassemble EIS1600 file(s) into MIU file(s).
-----
Give a single EIS1600 file as input
or 
Run without input arg to batch process all double-checked EIS1600 files from the AUTOREPORT.
'''
    )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 file to process',
            action=CheckFileEndingAction
    )
    args = arg_parser.parse_args()

    verbose = args.verbose

    if args.input:
        infile = './' + args.input
        out_path = get_path_to_other_repo(infile, 'MIU')
        print(f'Disassemble {infile}')
        disassemble_text(infile, out_path, verbose)
        infiles = [infile.split('/')[-1]]
        path = out_path.split('data')[0]
        write_to_readme(path, infiles, '# Texts disassembled into MIU files\n')
    else:
        input_dir = TEXT_REPO
        out_path = get_path_to_other_repo(input_dir, 'MIU')

        print(f'Disassemble double-checked EIS1600 files from the AUTOREPORT')
        files_list = read_files_from_autoreport(input_dir)

        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'EIS1600')
        if not infiles:
            print('There are no EIS1600 files to process')
            exit()

        if verbose:
            for infile in tqdm(infiles):
                try:
                    disassemble_text(infile, out_path, verbose)
                except Exception as e:
                    print(infile, e)
        else:
            res = []
            res += p_uimap(partial(disassemble_text, out_path=out_path), infiles)

        path = out_path.split('data')[0]
        write_to_readme(path, infiles, '# Texts disassembled into MIU files\n')

    print('Done')



from re import compile

from openiti.helper.ara import denormalize

from eis1600.helper.ar_normalization import normalize_dict


TOPONYM_CATEGORIES = {
        'ولد': 'B', 'مولد': 'B',
        'مات': 'D', 'موت': 'D', 'توفي': 'D', 'وفاة': 'D',
        'سمع': 'K', 'روى': 'K', 'روا': 'K', 'قرا': 'K', 'اجاز': 'K', 'حدث': 'K',
        'لقي': 'M',
        'استقر': 'O', 'انفصل': 'O', 'ولي': 'O', 'قاضي': 'O', 'نائب': 'O', 'أعمال': 'O',
        'حج': 'P',
        'سكن': 'R', 'نزل': 'R', 'نزيل': 'R', 'من اهل': 'R', 'استوطن': 'R', 'كان من': 'R', 'نشأ': 'R'
}
TOPONYM_CATEGORIES_NOR = normalize_dict(TOPONYM_CATEGORIES)

AR_TOPONYM_CATEGORIES = '|'.join([denormalize(key) for key in TOPONYM_CATEGORIES.keys()])
TOPONYM_CATEGORY_PATTERN = compile(r'\s[وف]?(?P<topo_category>' + AR_TOPONYM_CATEGORIES + r')')



from typing import List
import string

from eis1600.nlp.cameltools import lemmatize_and_tag_ner, CamelToolsModels


def annotate_miu_text(df):
    lemmas, ner_tags, pos_tags = ['_'], ['_'], ['_']
    section_id, temp_tokens = None, []
    for entry in list(zip(df['SECTIONS'].to_list(), df['TOKENS'].fillna('').to_list()))[1:]:
        _section, _token = entry[0], entry[1]
        if _section is not None:
            # Start a new section
            if len(temp_tokens) > 0:
                # 1. process the previous section
                _labels = lemmatize_and_tag_ner(temp_tokens)
                _, _ner_tags, _lemmas, _dediac_lemmas, _pos_tags = zip(*_labels)
                ner_tags.extend(_ner_tags)
                lemmas.extend(_dediac_lemmas)
                pos_tags.extend(_pos_tags)

                # 2. reset variables
                section_id, temp_tokens = None, []

        token = _token if _token not in ['', None] else '_'
        temp_tokens.append(token)

    if len(temp_tokens) > 0:
        _labels = lemmatize_and_tag_ner(temp_tokens)
        _, _ner_tags, _lemmas, _dediac_lemmas, _pos_tags = zip(*_labels)
        ner_tags.extend(_ner_tags)
        lemmas.extend(_dediac_lemmas)
        pos_tags.extend(_pos_tags)
    return ner_tags, lemmas, pos_tags


def camel2md(labels: list) -> List[str]:
    default_str = ''

    converted_tokens, temp_tokens, temp_class = [], [], None
    for _label in labels:
        if _label is None:
            converted_tokens.append(default_str)
        else:
            # Check if the first letter of the label is 'o' or 'B' a Begining of an NE
            if _label[0] in ['O', 'B', '_']:
                if len(temp_tokens) > 0 and temp_class is not None:
                    converted_tokens.append(f"{temp_class}{len(temp_tokens)}")  # e.g. ÜP3
                    converted_tokens.extend([default_str] * (len(temp_tokens) - 1))
                    # reset temp variables
                    temp_tokens, temp_class = [], None

                if _label in ['O', '_']:
                    converted_tokens.append(default_str)
                else:
                    temp_tokens.append(_label)
                    temp_class = _label[2:]

            else:
                if temp_class is not None:
                    temp_tokens.append(_label)
                else:
                    converted_tokens.append(default_str)
    if len(temp_tokens) > 0 and temp_class is not None:
        converted_tokens.append(f"{temp_class}{len(temp_tokens)}")
        converted_tokens.extend([default_str] * (len(temp_tokens) - 1))
    return converted_tokens

def camel2md_as_list(labels: list) -> List[str]:
    default_str = ''
    types_mapping = {
        'LOC': 'ÜT',
        'PERS': 'ÜP'
    }
    converted_tokens, temp_tokens, temp_class = [], [], None
    for _label in labels:
        if _label is None:
            converted_tokens.append(default_str)
        else:
            # Check if the first letter of the label is 'o' or 'B' a Begining of an NE
            if _label[0] in ['O', 'B', '_']:
                if len(temp_tokens) > 0 and temp_class is not None:
                    converted_tokens.append(f"{types_mapping.get(temp_class, 'ÜM')}{len(temp_tokens)}")  # e.g. ÜP3
                    converted_tokens.extend([default_str] * (len(temp_tokens) - 1))
                    # reset temp variables
                    temp_tokens, temp_class = [], None

                if _label in ['O', '_']:
                    converted_tokens.append(default_str)
                else:
                    temp_tokens.append(_label)
                    temp_class = _label[2:]

            else:
                if temp_class is not None:
                    temp_tokens.append(_label)
                else:
                    converted_tokens.append(default_str)
    if len(temp_tokens) > 0 and temp_class is not None:
        converted_tokens.append(f"{types_mapping.get(temp_class, 'ÜM')}{len(temp_tokens)}")
        converted_tokens.extend([default_str] * (len(temp_tokens) - 1))
    return converted_tokens


def insert_nasab_tag(df) -> list:
    tokens = df['TOKENS'].fillna('').to_list()
    nasab_tagger = CamelToolsModels.getNasabModel()
    shortend_list_of_tokens = tokens[1:]
    __shortend_list_limit = 120
    if len(tokens) > __shortend_list_limit:
        shortend_list_of_tokens = tokens[1:__shortend_list_limit]
        # print(tokens[:__shortend_list_limit], tokens)
        # for token in tokens[1:__shortend_list_limit]:
        #     if token is not None:
        #         shortend_list_of_tokens.append(token)
        #     else:
        #         shortend_list_of_tokens.append("_")
    nasab_labels = nasab_tagger.predict_sentence(shortend_list_of_tokens)
    punct = "..،_" + string.punctuation
    nasab = ['_']
    nasab_started = False
    for token, label in zip(shortend_list_of_tokens, nasab_labels):
        if label == "B-NASAB":
            # Start a new NASAB
            nasab.append("BNASAB")
            nasab_started = True
        else:
            if label == "I-NASAB":
                nasab.append('')
            else:
                if nasab_started:
                    if token.strip() in punct:
                        nasab[-1] = "ENASAB"
                        nasab.append("")
                    else:
                        nasab.append("ENASAB")
                    nasab_started = False
                else:
                    nasab.append('')
    if nasab_started:
        nasab[-1] = "ENASAB"
    # merge the shortend list
    if len(tokens) > __shortend_list_limit:
        nasab.extend([''] * (len(tokens) - __shortend_list_limit))

    return nasab


def insert_onomastic_tags(df):
    onomastic_tagger = CamelToolsModels.getOnomasticModel()
    onomastic_tags = [None] * len(df['TOKENS'])
    start_nasab_id, end_nasab_id = -1, -1

    # Find BNASAB & ENASAB
    for idx, tag in enumerate(df['NASAB_TAGS'].to_list()):
        if "BNASAB" == tag:
            start_nasab_id = idx
        elif "ENASAB" == tag:
            end_nasab_id = idx
            break

    if start_nasab_id > 0 and end_nasab_id > start_nasab_id:
        nasab_tokens = df['TOKENS'].to_list()[start_nasab_id:end_nasab_id]
        onomastic_labels = onomastic_tagger.predict_sentence(nasab_tokens)
        ono_tags = camel2md(onomastic_labels)

        for i, tag in enumerate(ono_tags):
            onomastic_tags[start_nasab_id + i] = tag

    df['ONONMASTIC_TAGS'] = onomastic_tags

    return df['ONONMASTIC_TAGS']


from sys import argv, exit
from os.path import isfile, splitext
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from multiprocessing import Pool

from eis1600.helper.repo import get_files_from_eis1600_dir, get_path_to_other_repo, read_files_from_readme
from eis1600.miu.methods import reassemble_text


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.IDs':
                parser.error('You need to input an IDs file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
        prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
        description='''Script to reassemble EIS1600 file(s) from MIU file(s).
-----
Give a single IDs file as input
or 
Use -e <EIS1600_repo> to batch process all files in the EIS1600 directory.
'''
        )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument(
        '-e', '--eis1600_repo', type=str,
        help='Takes a path to the EIS1600 file repo and batch processes all files'
        )
    arg_parser.add_argument(
        'input', type=str, nargs='?',
        help='EIS1600 file to process',
        action=CheckFileEndingAction
        )
    args = arg_parser.parse_args()

    verbose = args.verbose

    if args.input:
        infile = './' + args.input
        out_path = get_path_to_other_repo(infile, 'TEXT')
        print(f'Reassemble {infile}')
        reassemble_text(infile, out_path, verbose)
    elif args.eis1600_repo:
        input_dir = args.eis1600_repo
        if not input_dir[-1] == '/':
            input_dir += '/'
        out_path = get_path_to_other_repo(input_dir, 'TEXT')

        print(f'Reassemble EIS1600 files from the EIS1600 repo')
        files_list = read_files_from_readme(input_dir, '# Texts disassembled into MIU files\n')
        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'IDs')
        if not infiles:
            print('There are no IDs files to process')
            exit()

        params = [(infile, out_path, verbose) for infile in infiles]

        with Pool() as p:
            p.starmap_async(reassemble_text, params).get()
    else:
        print(
                'Pass in a <uri.IDs> file to process a single file or use the -e option for batch processing'
        )
        exit()

    print('Done')



from typing import List

from openiti.helper.ara import normalize_ara_heavy


def normalize_dict(o_set: dict) -> dict:
    n_set = {}

    for key, val in o_set.items():
        n_set[normalize_ara_heavy(key)] = val

    return n_set


def denormalize_list(o_list: List[str]) -> List[str]:
    # TODO evaluate if this is enough
    n_list = []
    for elem in o_list:
        tmp = [elem]
        if elem.startswith(('أ', 'ٱ', 'آ', 'إ')):
            # alifs
            tmp.append('ا' + elem[1:])
        if elem.endswith(('يء', 'ىء', 'ؤ', 'ئ')):
            # hamzas
            tmp.append([var[:-1] + 'ء' for var in tmp])
        if elem.endswith('ى'):
            # alif maqsura
            tmp.extend([var[:-1] + 'ي' for var in tmp])
        if elem.endswith('ي'):
            # alif maqsura
            tmp.extend([var[:-1] + 'ى' for var in tmp])
        if elem.endswith('ة'):
            # ta marbuta
            tmp.extend([var[:-1] + 'ه' for var in tmp])

        n_list.extend(tmp)

    return n_list



from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from pathlib import Path
import jsonpickle
from p_tqdm import p_uimap

from eis1600.processing.preprocessing import get_yml
from eis1600.helper.repo import MC_REPO, TRAINING_DATA_REPO


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to generate JSON from MIU YAMLHeaders.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    args = arg_parser.parse_args()

    debug = args.debug
    with open(TRAINING_DATA_REPO + 'gold_standard.txt', 'r', encoding='utf-8') as fh:
        files_txt = fh.read().splitlines()
    infiles = [TRAINING_DATA_REPO + 'gold_standard_nasab/' + file for file in files_txt if Path(
            TRAINING_DATA_REPO + 'gold_standard_nasab/' + file
    ).exists()]

    res = []
    if debug:
        for file in infiles[:10]:
            print(file)
            res.append(get_yml(file))
    else:
        res = p_uimap(get_yml, infiles)

    yml_dict = []
    for path, yml in res:
        key = path.replace(TRAINING_DATA_REPO + 'gold_standard_nasab/', '').replace('.EIS1600', '')
        author, text, version, UID = key.split('.')
        yml_init = {'author': author, 'text': text, 'version': version, 'UID': UID}
        yml_dict.append(yml.to_json(yml_init))

    with open(MC_REPO + 'masterchronicleapp/src/data.json', 'w', encoding='utf-8') as fh:
        jsonpickle.set_encoder_options('json', indent=4, ensure_ascii=False)
        json_str = jsonpickle.encode(yml_dict, unpicklable=False)
        fh.write(json_str)

    print('Done')



from os import makedirs
from os.path import dirname
from pathlib import Path
from typing import Optional
from numpy import nan
from pandas import Series
from pandas import DataFrame, notna

from eis1600.gazetteers.Onomastics import Onomastics
from eis1600.gazetteers.Toponyms import Toponyms
from eis1600.helper.logging import setup_logger
from eis1600.helper.repo import GAZETTEERS_REPO
from eis1600.miu.YAMLHandler import YAMLHandler
from eis1600.onomastics.re_pattern import ABI, ABU, BN_BNT, CRF_PATTERN, IBN_IBNA, SHR_PATTERN, SPELLING, UMM
from eis1600.processing.postprocessing import write_updated_miu_to_file
from eis1600.processing.preprocessing import get_tokens_and_tags, get_yml_and_miu_df

__log_filename = GAZETTEERS_REPO + 'logs/nasab_unknown.log'
makedirs(dirname(__log_filename), exist_ok=True)
LOGGER_NASAB_UNKNOWN = setup_logger('nasab_unknown', __log_filename)


def get_nas(text: str) -> str:
    """Add the list of forefathers to the YAMLHeader and return nasab part with tagged NAS and manipulated so it is
    ignored for further onomastic analysis (all elements which refere to ancestors are filtered here).

    :param str text: nasab str.
    :return str: nasab str with tagged and manipulated nas elements.
    """
    og = Onomastics.instance()

    text_mnpld = ABU.sub(' ابو_', text)
    text_mnpld = UMM.sub(' ام_', text_mnpld)
    text_mnpld = ABI.sub(' ابي_', text_mnpld)
    text_mnpld = IBN_IBNA.sub(r' \g<1>', text_mnpld)

    m = og.get_ngrams_regex().search(text_mnpld)
    while m:
        text_mnpld = text_mnpld[:m.start()] + m.group(1) + m.group(2).replace(' ', '_') + text_mnpld[m.end():]
        m = og.get_ngrams_regex().search(text_mnpld, m.end())

    m_bn = list(BN_BNT.finditer(text_mnpld))
    if m_bn:
        start = m_bn[0].start()
        end = m_bn[-1].end()
        pos_abu = text_mnpld[start:end].find(' ابو_')
        pos_umm = text_mnpld[start:end].find(' ام_')
        pos = max(pos_abu, pos_umm)
        if pos > 1:
            end = min(end, start + pos - 1)
        last_ancestor = text_mnpld[end + 1:].find(' ')
        if last_ancestor == -1:
            ancestors = BN_BNT.split(text_mnpld[start:])
        else:
            ancestors = BN_BNT.split(text_mnpld[start:end + 1 + last_ancestor])
        if ancestors[0] == '':
            ancestors = ancestors[1:]
        nas_w_tags = ''
        # Insert NAS tags
        for elem in ancestors:
            if not elem.startswith('بن'):
                num_tokens = len(elem.replace('_', ' ').split())
                nas_w_tags += ' ÜNAS' + str(max(1, num_tokens))
            nas_w_tags += ' ' + elem
        # Connect NAS elements with _ so they are ignored in further analysis
        if last_ancestor == -1:
            text_w_tags = text_mnpld[:start + 1].replace('_', ' ') + \
                          nas_w_tags[1:].replace(' ', '_')
        else:
            text_w_tags = text_mnpld[:start + 1].replace('_', ' ') + \
                          nas_w_tags[1:].replace(' ', '_') + \
                          text_mnpld[end + 1 + last_ancestor:].replace('_', ' ')

        return text_w_tags

    return text_mnpld


def tag_nasab(text: str) -> str:
    """Annotate the nasab part of the MIU.

    :param str text: nasab part of the MIU as one single string with tagged and filtered NAS elements (NAS is
    connected with '_' and therefore does not match with the gazetteer).
    :return str: the nasab part pf the MIU which contains also the tags in front of the recognized elements,
    '_' are removed.
    """
    og = Onomastics.instance()
    tg = Toponyms.instance()
    text_mnpld = text.replace(' بن ', ' بن_')
    # for m in BANU_BANI.finditer(text_mnpld):
    #     print(f'{m.group(0)} last: {text_mnpld[m.end()]}')
    text_mnpld = text_mnpld.replace('من ولد ', 'من_ولد_')
    m = og.get_ngrams_regex().search(text_mnpld)
    while m:
        tag = og.get_ngram_tag(m.group(2))
        pos = m.start()
        end = m.end()
        if m.group(1) == ' ':
            pos += 1
        if tag.startswith('ÜEXP'):
            if CRF_PATTERN.search(m.group(2)) or SHR_PATTERN.search(m.group(2)):
                tag = 'ÜSHR'
                if text_mnpld[end:].startswith('ابن '):
                    tag += '2 '
                    text_mnpld = text_mnpld[:end] + text_mnpld[end + 1:].replace(' ', '_', 1)
                else:
                    tag += '1 '
                if m.group(2).endswith(' ب'):
                    end -= 1

                text_mnpld = text_mnpld[:end] + tag + text_mnpld[end:]
                text_mnpld = text_mnpld[:pos] + \
                             text_mnpld[pos + 1:end + len(tag) + 1].replace(' ', '_') + \
                             text_mnpld[end + len(tag) + 1:]
            # else:
            #     tag = ''
            #     print(f'start: {text_mnpld[pos:end]} end: {text_mnpld[end:]}')
        else:
            text_mnpld = text_mnpld[:pos] + tag + text_mnpld[pos:]

        end += len(tag)

        m = og.get_ngrams_regex().search(text_mnpld, end)

    # Log unidentified tokens as uni- and bi-grams
    filtered = text_mnpld
    m = og.get_ngrams_regex().search(filtered)
    while m:
        filtered = filtered[:m.start()] + m.group(1) + m.group(2).replace(' ', '_') + filtered[m.end():]
        m = og.get_ngrams_regex().search(filtered, m.end())
    tokens = [token for token in filtered.split() if not token.startswith('Ü')]
    unknown_uni = [t for t in tokens if not ('_' in t or t in og.total() + tg.total())]
    prev = None
    unknown_bi = []
    for t in tokens:
        if not prev and not ('_' in t or t in og.total() + tg.total() + ['بن', 'بنت']):
            prev = t
        else:
            if not ('_' in t or t in og.total() + tg.total() + ['بن', 'بنت']):
                unknown_bi.append(prev + ' ' + t)
                prev = t
            else:
                prev = None
    if unknown_uni or unknown_bi:
        LOGGER_NASAB_UNKNOWN.info('\n'.join(unknown_uni + unknown_bi))

    return text_mnpld.replace('_', ' ')


def tag_spelling(text: str) -> str:
    """Tags spelling information which is stated in the nasab part of the MIU text.

    Spelling is detected when two elements of the spelling gazetteer are found successively.
    :param str text: the nasab part as one string.
    :return str: nasab part as one string including tags for found spelling.
    """
    text_updated = text
    m = SPELLING.search(text_updated)
    while m:
        tag = 'ÜSPL' + str(len(m.group(0).split())) + ' '
        pos = m.start()
        text_updated = text_updated[:pos] + tag + text_updated[pos:]

        m = SPELLING.search(text_updated, m.end() + len(tag))

    return text_updated


def nasab_annotate_miu(
        df: DataFrame,
        yml_handler: YAMLHandler,
        file: str,
        test: Optional[bool] = False
) -> Series:
    """Onomastic analysis of the nasab part of the MIU.

    :param DataFrame df: DataFrame of the MIU.
    :param YAMLHandler yml_handler: YAMLHandler of the MIU.
    :param Path file: the MIU which was opened.
    :param bool test: run on test data set, optional.
    :return Series: a series of the same length as the df containing the nasab tags corresponding to the tokens.
    """
    if not yml_handler.is_bio():
        return Series([nan] * len(df))

    s_notna = df['TAGS_LISTS'].loc[df['TAGS_LISTS'].notna()].apply(lambda tag_list: ','.join(tag_list))
    try:
        if test:
            # training_data batch uses old NASAB tag, not new ENASAB tag
            idx = s_notna.loc[s_notna.str.contains('NASAB')].index[0]
        else:
            idx = s_notna.loc[s_notna.str.contains('ENASAB')].index[0]
    except IndexError:
        print(
                f'MIU does not have a ENASAB tag, most likely the MIU is not a biography but a cross reference and '
                f'wrongly labeled. '
                f'Check:\n{file}'
        )
        return Series([nan] * len(df))

    # The following blacklisted elements are considered noise in the text data and are therefore ignored but kept
    blacklist = ['(', ')', '[', ']', '"', "'", '.', '،', '؟', '!', ':', '؛', ',', ';', '?', '|']
    nasab_idx = df.loc[df['TOKENS'].notna() & ~df['TOKENS'].isin(blacklist) & df.index.isin(df.iloc[:idx].index)].index

    text = ' '.join(df['TOKENS'].loc[nasab_idx])

    tagged_spelling = tag_spelling(text)
    ar_tokens, tags = get_tokens_and_tags(tagged_spelling)
    df.loc[nasab_idx, 'ONONMASTIC_TAGS'] = tags

    count = 0
    spl_idcs = []
    for row in df.loc[nasab_idx].itertuples():
        if notna(row[4]):
            count = int(row[4][-1])
        if count > 0:
            count -= 1
            spl_idcs.append(row[0])

    nasab_idx = df.loc[nasab_idx.difference(spl_idcs)].index
    text = ' '.join(df['TOKENS'].loc[nasab_idx])

    text_w_mnpld_nas = get_nas(text)
    tagged_onomastics = tag_nasab(text_w_mnpld_nas)
    ar_tokens, tags = get_tokens_and_tags(tagged_onomastics)
    df.loc[nasab_idx, 'ONONMASTIC_TAGS'] = tags

    if idx != len(df):
        # TODO make NASAB stay on same line
        df.loc[idx - 1, 'NASAB_END'] = 'NASAB'

    return df['ONONMASTIC_TAGS'].to_list()


def nasab_annotation(file: str, test: Optional[bool] = False):
    """Helper to run onomastic annotation standalone as cmdline script.

    :param str file: Path of the miu file to annotate.
    :param bool test: Indicating if the script is run with test data, defaults to false.
    """
    with open(file, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object)
    if test:
        # Only used if run on training_data batch because this information is missing there
        if '$' not in df.iloc[0]['SECTIONS'] or '$$$' in df.iloc[0]['SECTIONS'] or not yml_handler.is_reviewed():
            df['NASAB_TAGS'] = Series([nan] * len(df))
        else:
            yml_handler.set_category('$')
            df['NASAB_TAGS'] = nasab_annotate_miu(df, yml_handler, file, test)
    else:
        # Run on new data batch
        df['NASAB_TAGS'] = nasab_annotate_miu(df, yml_handler, file, test)
    yml_handler.unset_reviewed()

    if test:
        output_path = str(file).replace('gold_standard', 'gold_standard_nasab')
    else:
        output_path = str(file)

    with open(output_path, 'w', encoding='utf-8') as out_file_object:
        write_updated_miu_to_file(
                out_file_object, yml_handler, df[['SECTIONS', 'TOKENS', 'TAGS_LISTS', 'NASAB_TAGS']]
        )



from eis1600.miu.YAMLHandler import YAMLHandler

from eis1600.miu.yml_handling import extract_yml_header_and_text
from typing import Iterator, List, TextIO, Tuple, Union

from pandas import DataFrame, options

from camel_tools.tokenizers.word import simple_word_tokenize
from eis1600.helper.markdown_patterns import MIU_TAG_PATTERN, SECTION_PATTERN, SECTION_SPLITTER_PATTERN, TAG_PATTERN


options.mode.chained_assignment = None


def get_tokens_and_tags(tagged_text: str) -> Tuple[List[Union[str, None]], List[Union[str, None]]]:
    """Splits the annotated text into two lists of the same length, one containing the tokens, the other one the tags

    :param str tagged_text: the annotated text as a single str.
    :return List[str], List[str]: two lists, first contains the arabic tokens, the other one the tags.
    """
    tokens = simple_word_tokenize(tagged_text)
    ar_tokens, tags = [], []
    tag = None
    for t in tokens:
        if TAG_PATTERN.match(t):
            tag = t
        else:
            ar_tokens.append(t)
            tags.append(tag)
            tag = None

    return ar_tokens, tags


def tokenize_miu_text(text: str) -> Iterator[Tuple[Union[str, None], Union[str, None], List[Union[str, None]]]]:
    """Returns the MIU text as zip object of three sparse columns: sections, tokens, lists of tags.

    Takes an MIU text and returns a zip object of three sparse columns: sections, tokens, lists of tags. Elements can
    be None because of sparsity.
    :param text: MIU text content to process.
    :return Iterator: Returns a zip object containing three sparse columns: sections, tokens, lists of tags. Elements
    can be None because of sparsity.
    """
    text_and_heading = MIU_TAG_PATTERN.split(text)
    # The indices are connected to the number of capturing group in MIU_TAG_PATTERN
    heading = text_and_heading[1]
    text_iter = SECTION_SPLITTER_PATTERN.split(text_and_heading[4][:-2]).__iter__()
    paragraph = next(text_iter)

    sections, ar_tokens, tags = [heading], [None], [None]
    section = None

    # First item in text_iter is an empty string if there are multiple paragraphs therefore test for None
    while paragraph is not None:
        if SECTION_PATTERN.fullmatch(paragraph):
            section = paragraph
        else:
            # Encode \n with NEWLINE as they will be removed by the simple_word_tokenize method
            # NEWLINE is treated like a tag
            text_wo_new_lines = paragraph.replace('\n_ء_', ' NEWLINE ')
            text_wo_new_lines = text_wo_new_lines.replace('\n', ' NEWLINE ')
            text_wo_new_lines = text_wo_new_lines.replace('%~%', 'HEMISTICH')
            tokens = simple_word_tokenize(text_wo_new_lines)
            tag = None
            for t in tokens:
                if TAG_PATTERN.match(t):
                    if not t.startswith('Ü'):
                        # Do not add automated tags to the list - they come from the csv anyway
                        # There might be multiple tags in front of a token - Page, NEWLINE, NER tag, ...
                        if tag:
                            tag.append(t)
                        else:
                            tag = [t]
                else:
                    sections.append(section)
                    section = None
                    ar_tokens.append(t)
                    tags.append(tag)
                    tag = None
            if tag:
                sections.append(section)
                section = None
                ar_tokens.append('')
                tags.append(tag)

        paragraph = next(text_iter, None)

    return zip(sections, ar_tokens, tags)


def get_yml_and_miu_df(miu_file_object: TextIO) -> Tuple[YAMLHandler, DataFrame]:
    """Returns YAMLHandler instance and MIU as a DataFrame containing the columns 'SECTIONS', 'TOKENS', 'TAGS_LISTS'.

    :param TextIO miu_file_object: File object of the MIU file.
    :return Tuple[YAMLHandler, DataFrame]: YAMLHandler and DataFrame containing the columns 'SECTIONS', 'TOKENS',
    'TAGS_LISTS'.
    """
    yml_str, text = extract_yml_header_and_text(miu_file_object, False)
    yml_handler = YAMLHandler().from_yml_str(yml_str)
    zipped = tokenize_miu_text(text)
    df = DataFrame(zipped, columns=['SECTIONS', 'TOKENS', 'TAGS_LISTS'])

    df.mask(df == '', inplace=True)

    return yml_handler, df


def get_yml(path: str) -> Tuple[str, YAMLHandler]:
    with open(path, 'r', encoding='utf-8') as miu_file_object:
        yml_str, text = extract_yml_header_and_text(miu_file_object, False)
    return path, YAMLHandler().from_yml_str(yml_str)



from typing import Literal


class Date:
    """

    :param int year: year.
    :param int length: number of tokens which make up the date expression.
    :param Literal['B', 'D', 'K', 'H', 'P', 'L'] category:
    """
    def __init__(self, year: int, length: int, category: Literal['B', 'D', 'K', 'H', 'P', 'L', 'X'] = 'X'):
        self.year = year
        self.length = length
        self.category = category

    def __eq__(self, other):
        return self.year == other.year and self.length == other.length and self.category == other.category

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    def get_tag(self):
        tag = 'ÜY'
        tag += str(self.length)
        tag += self.category

        year = str(self.year)
        while len(year) < 4:
            year = '0' + year

        tag += year

        return tag + 'Y '



from typing import Iterator, List, TextIO, Tuple, Union

from os import path
from pandas import DataFrame, notna
from camel_tools.utils.charsets import UNICODE_PUNCT_CHARSET

from eis1600.helper.markdown_patterns import ENTITY_TAGS_PATTERN
from eis1600.miu.YAMLHandler import YAMLHandler
from eis1600.miu.yml_handling import add_annotated_entities_to_yml


def get_text_with_annotation_only(
        text_and_tags: Union[Iterator[Tuple[Union[str, None], str, Union[List[str], None]]], DataFrame]
) -> str:
    """Returns the MIU text only with annotation tags, not page tags and section tags.

    Returns the MIU text only with annotation tags contained in the list of tags. Tags are inserted BEFORE the token.
    Section headers and other tags - like page tags - are ignored.
    :param Iterator[Tuple[Union[str, None], str, Union[List[str], None]]] text_and_tags: zip object containing three
    sparse columns: sections, tokens, lists of tags.
    :return str: The MIU text with annotation only.
    """
    if type(text_and_tags) is DataFrame:
        text_and_tags_iter = text_and_tags.itertuples(index=False)
    else:
        text_and_tags_iter = text_and_tags.__iter__()
    next(text_and_tags_iter)
    text_with_annotation_only = ''
    for section, token, tags in text_and_tags_iter:
        if isinstance(tags, list):
            entity_tags = [tag for tag in tags if ENTITY_TAGS_PATTERN.fullmatch(tag)]
            text_with_annotation_only += ' ' + ' '.join(entity_tags)
        if notna(token):
            text_with_annotation_only += ' ' + token

    return text_with_annotation_only


def reconstruct_miu_text_with_tags(
        text_and_tags: Union[Iterator[Tuple[Union[str, None], str, Union[List[str], None]]], DataFrame]
) -> str:
    """Reconstruct the MIU text from a zip object containing three columns: sections, tokens, lists of tags.

    Reconstructs the MIU text with the tags contained in the list of tags. Tags are inserted BEFORE the token.
    Section headers are inserted after an empty line ('\n\n'), followed by the text on the next line.
    :param Iterator[Tuple[Union[str, None], str, Union[List[str], None]]] text_and_tags: zip object containing three
    sparse columns: sections, tokens, lists of tags.
    :return str: The reconstructed MIU text containing all the tags.
    """
    if type(text_and_tags) is DataFrame:
        text_and_tags_iter = text_and_tags.itertuples(index=False)
    else:
        text_and_tags_iter = text_and_tags.__iter__()
    heading, _, _ = next(text_and_tags_iter)
    reconstructed_text = heading
    # TODO NASAB tag after token
    for section, token, tags in text_and_tags_iter:
        if notna(section):
            reconstructed_text += '\n\n' + section + '\n_ء_'
        if isinstance(tags, list):
            reconstructed_text += ' ' + ' '.join(tags)
        if notna(token):
            if token in UNICODE_PUNCT_CHARSET:
                reconstructed_text += token
            else:
                reconstructed_text += ' ' + token

    reconstructed_text += '\n\n'
    reconstructed_text = reconstructed_text.replace(' NEWLINE ', '\n_ء_ ')
    reconstructed_text = reconstructed_text.replace('HEMISTICH', '%~%')
    return reconstructed_text


def merge_tagslists(lst1, lst2):
    if isinstance(lst1, list):
        if notna(lst2):
            lst1.append(lst2)
    else:
        if notna(lst2):
            lst1 = [lst2]
    return lst1


def write_updated_miu_to_file(
        miu_file_object: TextIO,
        yml_handler: YAMLHandler,
        df: DataFrame,
) -> None:
    """Write MIU file with annotations and populated YAML header.

    :param TextIO miu_file_object: Path to the MIU file to write
    :param YAMLHandler yml_handler: The YAMLHandler of the MIU.
    :param DataFrame df: df containing the columns ['SECTIONS', 'TOKENS', 'TAGS_LISTS'] and optional 'ÜTAGS_LISTS'.
    :return None:
    """
    if not yml_handler.is_reviewed():
        columns_of_automated_tags = ['NER_TAGS', 'DATE_TAGS', 'NASAB_TAGS', 'ONONMASTIC_TAGS']
        df['ÜTAGS'] = df['TAGS_LISTS']
        for col in columns_of_automated_tags:
            if col in df.columns:
                df['ÜTAGS'] = df.apply(lambda x: merge_tagslists(x['ÜTAGS'], x[col]), axis=1)
        df_subset = df[['SECTIONS', 'TOKENS', 'ÜTAGS']]
    else:
        df_subset = df[['SECTIONS', 'TOKENS', 'TAGS_LISTS']]

    text_with_tags = get_text_with_annotation_only(df_subset)
    add_annotated_entities_to_yml(text_with_tags, yml_handler, path.realpath(miu_file_object.name))
    updated_text = reconstruct_miu_text_with_tags(df_subset)

    miu_file_object.seek(0)
    miu_file_object.write(str(yml_handler) + updated_text)
    miu_file_object.truncate()



from itertools import combinations
from os import makedirs
from os.path import dirname, split, splitext
from typing import Dict, List, Optional, Set, TextIO, Tuple, Union

from eis1600.gazetteers.Toponyms import Toponyms
from eis1600.helper.EntityTags import EntityTags
from eis1600.helper.logging import setup_logger
from eis1600.helper.markdown_methods import get_yrs_tag_value
from eis1600.helper.markdown_patterns import ENTITY_TAGS_PATTERN, MIU_HEADER_PATTERN, NEWLINES_CROWD_PATTERN
from eis1600.helper.repo import GAZETTEERS_REPO
from eis1600.miu.HeadingTracker import HeadingTracker
from eis1600.miu.YAMLHandler import YAMLHandler


__log_filename_nasab = GAZETTEERS_REPO + 'logs/nasab_known.log'
makedirs(dirname(__log_filename_nasab), exist_ok=True)
LOGGER_NASAB_KNOWN = setup_logger('nasab_known', __log_filename_nasab)
LOGGER_TOPONYMS_UNKNOWN = setup_logger('toponyms_unknown', GAZETTEERS_REPO + 'logs/toponyms_unknown.log')


def create_yml_header(category: str, headings: Optional[HeadingTracker] = None) -> str:
    """Creates a YAML header for the current MIU file and returns it as yamlfied string.

    :param str category: Category of the entry.
    :param Type[HeadingsTracker] headings: HeadingTracker with the super elements of the current MIU, optional.
    :return str: YAML header for the current MIU.
    """
    yml_header = YAMLHandler()
    yml_header.set_category(category)
    if headings:
        yml_header.set_headings(headings)

    return yml_header.get_yamlfied()


def extract_yml_header_and_text(miu_file_object: TextIO, is_header: Optional[bool] = False) -> (str, str):
    """ Returns the YAML header and the text as a tuple from MIU file object.

    Splits the MIU file into a tuple of YAML header and text.
    :param TextIO miu_file_object: File object of the MIU file from which to extract YAML header and text.
    :param bool is_header: Indicates if the current MIU is the YAML header of the whole work and if so skips
    removing blank lines, defaults to False.
    :return (str, str): Tuple of the extracted YAML header and text.
    """
    text = ''
    miu_yml_header = ''
    for line in iter(miu_file_object):
        if MIU_HEADER_PATTERN.match(line):
            # Omit the #MIU#Header# line as it is only needed inside the MIU.EIS1600 file, but not in YMLDATA.yml
            next(miu_file_object)
            line = next(miu_file_object)
            miu_yml_header = ''
            while not MIU_HEADER_PATTERN.match(line):
                miu_yml_header += line
                line = next(miu_file_object)
            # Omit the empty line between the header content and the #MIU#Header# line
            miu_yml_header = miu_yml_header[:-2]
            # Skip empty line after #MIU#Header#
            next(miu_file_object)
        else:
            text += line
        # Replace new lines which separate YAML header from text
        if not is_header:
            text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)

    return miu_yml_header, text


def add_to_entities_dict(
        entities_dict: Dict, cat: str,
        entity: Union[str, Tuple[str, Union[int, str]], List[Tuple[str, str]], List[str], Tuple[int, str], Dict],
        tag: Optional[str] = None
) -> None:
    """Add a tagged entity to the respective list in the entities_dict.

    :param Dict entities_dict: Dict containing previous tagged entities.
    :param str cat: Category of the entity.
    :param Union[str|int] entity: Entity - might be int if entity is a date, otherwise str.
    :param str tag: Onomastic classification, used to differentiate between onomastic elements, optional.
    """
    cat = cat.lower() + 's'
    if tag:
        tag = tag.lower()
    if cat in entities_dict.keys():
        if cat == 'onomastics' and tag:
            if tag in entities_dict[cat].keys():
                entities_dict[cat][tag].append(entity)
            else:
                entities_dict[cat][tag] = [entity]
        else:
            entities_dict[cat].append(entity)
    else:
        if cat == 'onomastics' and tag:
            entities_dict[cat] = {}
            entities_dict[cat][tag] = [entity]
        elif isinstance(entity, list):
            entities_dict[cat] = entity
        else:
            entities_dict[cat] = [entity]


def add_annotated_entities_to_yml(
        text_with_tags: str,
        yml_handler: YAMLHandler,
        file_path: str,
) -> None:
    """Populates YAMLHeader with annotated entities.

    :param str text_with_tags: Text with inserted tags of the MIU.
    :param YAMLHandler yml_handler: YAMLHandler of the MIU.
    :param str file_path: Filename of the current MIU (used in error msg).
    """
    # We do not need to differentiate between automated and manual tags
    text_with_tags = text_with_tags.replace('Ü', '')
    tg = Toponyms.instance()
    entity_tags_df = EntityTags.instance().get_entity_tags_df()
    entities_dict = {}
    nas_dict = {}
    settlements_set: Set[str] = set()
    provinces_set: Set[str] = set()
    nas_counter = 0

    m = ENTITY_TAGS_PATTERN.search(text_with_tags)
    while m:
        tag = m.group('entity')
        length = int(m.group('length'))
        entity = ' '.join(text_with_tags[m.end():].split(maxsplit=length)[:length])

        cat = entity_tags_df.loc[entity_tags_df['TAG'].str.fullmatch(tag), 'CATEGORY'].iloc[0]
        if cat == 'DATE' or cat == 'AGE':
            try:
                val, e_cat = get_yrs_tag_value(m.group(0))
                add_to_entities_dict(entities_dict, cat, {'entity': entity, cat.lower(): val, 'cat': e_cat})
            except ValueError:
                print(f'Tag is neither year nor age: {m.group(0)}\nCheck: {file_path}')
                return
        elif cat == 'TOPONYM':
            place, tag, list_of_uris, list_of_provinces = tg.look_up_entity(entity)
            add_to_entities_dict(entities_dict, cat, {'entity': place, 'URI': tag})
            if len(list_of_uris) == 0:
                path, uri = split(file_path)
                uri, ext = splitext(uri)
                LOGGER_TOPONYMS_UNKNOWN.info(f'{uri},{entity}')
            else:
                if len(list_of_provinces) > 0:
                    settlements_set.update(list_of_uris)
                    provinces_set.update(list_of_provinces)
                else:
                    # if there is no province URI that means the toponym is a province and should therefore be added
                    # to provinces and not settlements
                    provinces_set.update(list_of_uris)
                if len(list_of_uris) > 1:
                    yml_handler.set_ambigious_toponyms()
        elif cat == 'ONOMASTIC':
            if tag.startswith('SHR') and entity.startswith('ب'):
                entity = entity[1:]
                add_to_entities_dict(entities_dict, cat, entity, tag)
            elif tag.startswith('NAS'):
                nas_dict['nas_' + str(nas_counter)] = entity
                nas_counter += 1
            add_to_entities_dict(entities_dict, cat, entity, tag)
            LOGGER_NASAB_KNOWN.info(f'{tag},{entity}')
        else:
            add_to_entities_dict(entities_dict, cat, entity, tag)

        m = ENTITY_TAGS_PATTERN.search(text_with_tags, m.end())

    if nas_dict != {}:
        if 'onomastics' in entities_dict.keys():
            entities_dict['onomastics']['nas'] = nas_dict
        else:
            entities_dict['onomastics'] = {'nas': nas_dict}

    if 'onomastics' in entities_dict.keys():
        # Sort dict by keys
        entities_dict['onomastics'] = dict(sorted(entities_dict.get('onomastics').items()))

    if settlements_set:
        entities_dict['settlements'] = list(settlements_set)
        entities_dict['edges_settlements'] = [[a, b] for a, b in combinations(settlements_set, 2) if a != b]
    if provinces_set:
        entities_dict['provinces'] = list(provinces_set)
        entities_dict['edges_provinces'] = [[a, b] for a, b in combinations(provinces_set, 2) if a != b]

    if entities_dict.get('dates'):
        dates = entities_dict.get('dates')
        birth_date = [d.get('date') for d in dates if d.get('cat') == 'B']
        death_date = [d.get('date') for d in dates if d.get('cat') == 'D']
        alternative_death_date = [d.get('date') for d in dates if d.get('cat') == 'B']
        if death_date:
            entities_dict['max_date'] = max(death_date)
        elif alternative_death_date:
            entities_dict['max_date'] = max(alternative_death_date)
        elif birth_date:
            entities_dict['max_date'] = min(birth_date) + 70

        if birth_date:
            entities_dict['min_date'] = min(birth_date)
        elif entities_dict.get('max_date'):
            entities_dict['min_date'] = entities_dict.get('max_date') - 70

    yml_handler.add_tagged_entities(entities_dict)



from pathlib import Path
from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from typing import Dict, Union
from re import compile

from json import dump
from p_tqdm import p_uimap

from eis1600.helper.repo import TRAINING_DATA_REPO
from eis1600.miu.methods import get_yml_and_miu_df


TOP_PATTERN = compile("T(?P<num_tokens>\d)(?P<category>[BDKMPR])")
categories = ["TOB", "TOD", "TOK", "TOM", "TOP", "TOR"]
BIO = ["B", "I"]

labels = [bi + "-" + c for c in categories for bi in BIO] + ["O"]
label_dict = {}

for i, label in enumerate(labels):
    label_dict[label] = i


def top_tag_to_bio(md_tag):
    m = TOP_PATTERN.search(md_tag)
    num_tokens = int(m.group("len"))
    cat = m.group("cat")

    bio_tags = []
    while len(bio_tags) < num_tokens:
        if len(bio_tags) == 0:
            bio_tags.append("B-TO" + cat)
        else:
            bio_tags.append("I-TO" + cat)

    return bio_tags


def md_to_bio(file: str) -> Union[Dict, None]:
    with open(file, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object)

    if not yml_handler.is_bio():
        return None

    s_notna = df['TAGS_LISTS'].loc[df['TAGS_LISTS'].notna()].apply(lambda tag_list: ','.join(tag_list))
    df_matches = s_notna.str.extract(TOP_PATTERN).dropna()

    if df_matches.empty:
        return None

    for index, row in df_matches.iterrows():
        processed_tokens = 0
        num_tokens = int(row['num_tokens'])
        while processed_tokens < num_tokens:
            if processed_tokens == 0:
                df.loc[index, 'BIO'] = 'B-TO' + row['category']
            else:
                df.loc[index + processed_tokens, 'BIO'] = 'I-TO' + row['category']

            processed_tokens += 1

    df["BIO"].loc[df["BIO"].isna()] = "O"
    df["BIO_IDS"] = df["BIO"].apply(lambda bio_tag: label_dict[bio_tag])

    idcs = df["TOKENS"].loc[df["TOKENS"].notna()].index

    return {
            "tokens": df["TOKENS"].loc[idcs].to_list(),
            "ner_tags": df["BIO_IDS"].loc[idcs].to_list(),
            "ner_classes": df["BIO"].loc[idcs].to_list()
    }


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate onomastic information in gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')

    args = arg_parser.parse_args()
    debug = args.debug

    with open(TRAINING_DATA_REPO + 'gold_standard.txt', 'r', encoding='utf-8') as fh:
        files_txt = fh.read().splitlines()

    infiles = [TRAINING_DATA_REPO + 'gold_standard_topo/' + file for file in files_txt if Path(
            TRAINING_DATA_REPO + 'gold_standard_topo/' + file).exists()]

    if debug:
        for file in infiles[:20]:
            print(file)
            md_to_bio(file)
    else:
        res = []
        res += p_uimap(md_to_bio, infiles)

        with open('toponyms_category_training_data.json', 'w', encoding='utf-8') as fh:
            dump([r for r in res if r is not None], fh, indent=4, ensure_ascii=False)

    print('Done')



from camel_tools.ner import NERecognizer
from camel_tools.tokenizers.word import simple_word_tokenize
from camel_tools.disambig.mle import MLEDisambiguator
from camel_tools.tagger.default import DefaultTagger
from camel_tools.utils.dediac import dediac_ar
from typing import Union


class CamelToolsModels:
    __pos_tagger = None
    __mled_disambiguator = None
    __lemmatizer = None
    __ner_tagger = None
    __nasab_tagger = None
    __onomastic_tagger = None
    __NASAB_MODEL_PATH = "EIS1600_Pretrained_Models/camelbert-ca-finetuned_nasab/"
    __NER_MODEL_PATH = "EIS1600_Pretrained_Models/camelbert-ca-finetuned_ner/"
    __ONOMASTIC_MODEL_PATH = "EIS1600_Pretrained_Models/camelbert-ca-finetuned_onomastic/"

    @staticmethod
    def getInstance():
        """ Static access method. """
        if CamelToolsModels.__ner_tagger is None or CamelToolsModels.__lemmatizer is None:
            CamelToolsModels()
        return CamelToolsModels.__ner_tagger, CamelToolsModels.__lemmatizer, CamelToolsModels.__pos_tagger

    @staticmethod
    def getNasabModel():
        """ Static access method. """
        if CamelToolsModels.__nasab_tagger is None:
            CamelToolsModels.__nasab_tagger = NERecognizer(CamelToolsModels.__NASAB_MODEL_PATH)
        return CamelToolsModels.__nasab_tagger

    @staticmethod
    def getOnomasticModel():
        """ Static access method. """
        if CamelToolsModels.__onomastic_tagger is None:
            CamelToolsModels.__onomastic_tagger = NERecognizer(CamelToolsModels.__ONOMASTIC_MODEL_PATH)
        return CamelToolsModels.__onomastic_tagger

    def __init__(self):
        """ Virtually private constructor. """
        if CamelToolsModels.__ner_tagger is not None:
            raise Exception("This class is a singleton!")
        else:
            CamelToolsModels.__mled_disambiguator = MLEDisambiguator.pretrained()
            CamelToolsModels.__lemmatizer = DefaultTagger(CamelToolsModels.__mled_disambiguator, 'lex')
            CamelToolsModels.__pos_tagger = DefaultTagger(CamelToolsModels.__mled_disambiguator, 'pos')
            CamelToolsModels.__ner_tagger = NERecognizer(CamelToolsModels.__NER_MODEL_PATH) # .pretrained


def lemmatize_and_tag_ner(tokens: Union[str, list]) -> list:
    """Lemmatize the text and annotate named-entities.

        Lemmatize the text and annotated named-entities using Camel Tools models.
        :param tokens: a  string or a list of tokens to be annotated
        """
    ner_tagger, lemmatizer, pos_tagger = CamelToolsModels.getInstance()
    # if tokens is a string, then tokenize it
    if isinstance(tokens, str):
        tokens = simple_word_tokenize(tokens)
    ner_labels = ner_tagger.predict_sentence(tokens)
    lemmas = lemmatizer.tag(tokens)
    pos_tags = pos_tagger.tag(tokens)
    dediac_lemmas = [dediac_ar(lemma) for lemma in lemmas]
    return list(zip(tokens, ner_labels, lemmas, dediac_lemmas, pos_tags))



from os.path import split

from typing import List, Optional, Tuple


def extract_categories(infile: str, verbose: Optional[bool] = None) -> Tuple[List[str], List[str]]:
    path, uri = split(infile)
    uri, ext = uri.split('.YAMLDATA.yml')

    if verbose:
        print(f'Check {uri}')
        print(infile)

    with open(infile, 'r', encoding='utf-8') as yml_data_fh:
        content = yml_data_fh.read()

    mius = content.split('#MIU#Header#End#')
    categories = []
    reviewers = []

    for miu in mius:
        cat = [line.lstrip('category   : ').strip('"') for line in miu.split('\n') if line.startswith('category')]
        categories.extend(cat)
        rev = [line.lstrip('reviewer   : ').strip('"') for line in miu.split('\n') if line.startswith('reviewer')]
        reviewers.extend(rev)

    return categories, reviewers



from typing import Optional
from sys import argv, exit
from os.path import isfile, split, splitext
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from multiprocessing import Pool

from eis1600.helper.repo import get_files_from_eis1600_dir, write_to_readme, read_files_from_readme
from eis1600.helper.markdown_patterns import BIO_CHR_TO_NEWLINE_PATTERN, HEADER_END_PATTERN, HEADING_OR_BIO_PATTERN, \
    MIU_LIGHT_OR_EIS1600_PATTERN, NEWLINES_CROWD_PATTERN, UID_PATTERN
from eis1600.markdown.UIDs import UIDs


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.EIS1600':
                parser.error('You need to input an EIS1600 file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def xx_update_uids(infile: str, verbose: Optional[bool] = False) -> None:
    """Updates a text with missing UIDs.

    :param str infile: Path of the file to update UIDs in.
    :param bool verbose: If True outputs a notification of the file which is currently processed, defaults to False.
    :return None:
    """

    path, ext = splitext(infile)
    outfile = path + '.EIS1600'
    path, uri = split(infile)

    if verbose:
        print(f'Update UIDs in {uri}')

    with open(infile, 'r', encoding='utf8') as infile_h:
        text = infile_h.read()

    text = BIO_CHR_TO_NEWLINE_PATTERN.sub(r'\1\n\2', text)
    header_and_text = HEADER_END_PATTERN.split(text)
    header = header_and_text[0] + header_and_text[1]
    text = header_and_text[2]
    text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)
    text = text.split('\n\n')
    text_updated = []

    used_ids = []

    for paragraph in text:
        if UID_PATTERN.match(paragraph):
            used_ids.append(int(UID_PATTERN.match(paragraph).group('UID')))

    uids = UIDs(used_ids)

    text_iter = text.__iter__()
    paragraph = next(text_iter, None)
    while paragraph is not None:
        next_p = next(text_iter, None)

        if paragraph:
            # Only do this is paragraph is not empty
            if HEADING_OR_BIO_PATTERN.match(paragraph):
                paragraph = paragraph.replace('#', f'_ء_#={uids.get_uid()}=')
                if next_p and not MIU_LIGHT_OR_EIS1600_PATTERN.match(next_p):
                    heading_and_text = paragraph.split('\n', 1)
                    if len(heading_and_text) > 1:
                        paragraph = heading_and_text[0] + f'\n\n_ء_={uids.get_uid()}= ::UNDEFINED:: ~\n' + \
                                    heading_and_text[1]
            elif not UID_PATTERN.match(paragraph):
                section_header = '' if paragraph.startswith('::') else '::UNDEFINED:: ~\n'
                paragraph = f'_ء_={uids.get_uid()}= {section_header}' + paragraph

            text_updated.append(paragraph)

        paragraph = next_p

    text = '\n\n'.join(text_updated)

    # reassemble text
    final = header + '\n\n' + text

    with open(outfile, 'w', encoding='utf8') as outfile_h:
        outfile_h.write(final)


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to insert UIDs in updated EIS1600 file(s).
-----
Give a single EIS1600 file as input
or 
Use -e <EIS1600_repo> to batch process all EIS1600 files in the EIS1600 directory which have not been processed yet.
'''
    )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument(
            '-e', '--eis1600_repo', type=str,
            help='Takes a path to the EIS1600 file repo and batch processes all files which have not been processed yet'
    )
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 file to process',
            action=CheckFileEndingAction
    )
    args = arg_parser.parse_args()

    verbose = args.verbose

    if args.input:
        infile = './' + args.input
        xx_update_uids(infile, verbose)
    elif args.eis1600_repo:
        input_dir = args.eis1600_repo
        if not input_dir[-1] == '/':
            input_dir += '/'

        print(
                f'Insert missing UIDs into checked files from the EIS1600 repo (only for EIS1600 files which have not '
                f'been updated yet)'
        )
        files_list = read_files_from_readme(
            input_dir, '# Texts which need to be updated (old process)\n', False
            )
        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'EIS1600')
        if not infiles:
            print(
                    'There are no more files to update'
            )
            exit()

        params = [(infile, verbose) for infile in infiles]

        with Pool() as p:
            p.starmap_async(xx_update_uids, params).get()

        write_to_readme(input_dir, infiles, '# Texts updated with missing UIDs\n')
    else:
        print(
                'Pass in a <uri.EIS1600> file to process a single file or use the -e option for batch processing'
        )
        exit()

    print('Done')



from __future__ import annotations
from copy import deepcopy

from typing import Dict, Optional


class HeadingTracker:
    """A class to keep track of the super elements of a MIU.

    This class keeps track of the headings on different levels to keep this information in the MIU YAML header. Some
    headings are empty stings
    :ivar str h1: Level 1 heading, optional.
    :ivar str h2: Level 2 heading, optional.
    :ivar str h3: Level 3 heading, optional.
    :ivar str h4: Level 4 heading, optional.
    """

    def __init__(self, headings_dict: Optional[Dict] = None) -> None:
        """Constructor which sets attributes to empty strings."""

        self.h1 = None
        self.h2 = None
        self.h3 = None
        self.h4 = None
        self.page_tag = None

        if headings_dict:
            for key, val in headings_dict.items():
                self.__setattr__(key, val)

    def __iter__(self):
        """Iterate over headings which are not None. Omits page_tag."""
        for key, val in self.__dict__.items():
            if key.startswith('h') and val is not None:
                yield key, val

    def get_curr_state(self) -> HeadingTracker:
        """Get current state of the tacker as deepcopy.

        Returns a deepcopy of the current state.
        :return HeadingTracker: Deepcopy of the current state of the tracker.
        """

        return deepcopy(self)

    def get_yamlfied(self) -> str:
        """Stringifies HeadingTracker in YAML format, only includes levels which are set.

        :return str: returns the HeadingTracker in YAML format as a string.
        """

        if self.h1 is None:
            return '[]'

        heading_tracker_str = '\n'
        heading_tracker_str += '    - h1    : \'' + self.h1 + '\'\n'
        if self.h2 is not None:
            heading_tracker_str += '    - h2    : \'' + self.h2 + '\'\n'
            if self.h3 is not None:
                heading_tracker_str += '    - h3    : \'' + self.h3 + '\'\n'
                if self.h4 is not None:
                    heading_tracker_str += '    - h4    : \'' + self.h4 + '\'\n'

        if self.page_tag:
            heading_tracker_str += '    - page_tag    : \'' + self.page_tag + '\''
        else:
            heading_tracker_str = heading_tracker_str[:-1]

        return heading_tracker_str

    def to_json(self) -> Dict:
        json_dict = {}
        for key, val in vars(self).items():
            if val is not None:
                json_dict[key] = val
        return json_dict

    def track_headings(self, level: int, heading: str) -> None:
        """Checks which of the levels changed and sets all sub levels to None (some headings are just an empty string).

        :param int level: The level of the heading indicated by the number of leading `|`.
        :param str heading: The new heading text for the given level.
        """

        if level == 1:
            self.h1 = heading
            self.h2 = None
            self.h3 = None
            self.h4 = None
        elif level == 2:
            self.h2 = heading
            self.h3 = None
            self.h4 = None
        elif level == 3:
            self.h3 = heading
            self.h4 = None
        else:
            self.h4 = heading

    def track_pages(self, page_tag: str) -> None:
        self.page_tag = page_tag.strip()

    def __repr__(self) -> str:
        return str(self.__dict__)

    def __str__(self) -> str:
        return self.get_yamlfied()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='eis1600',
      version='0.9.5',
      description='EIS1600 project tools and utilities',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/EIS1600/eis1600-pkg',
      author='Lisa Mischer',
      author_email='mischer.lisa@gmail.com',
      license='MIT License',
      packages=find_packages(include=['eis1600', 'eis1600.*'], exclude=['excluded']),
      package_data={'eis1600.gazetteers.data': ['*.csv'], 'eis1600.helper.data': ['*.csv']},
      entry_points={
          'console_scripts': [
                  'annotate_mius = eis1600.nlp.ner_annotate_mius:main [NER]',
                  'convert_mARkdown_to_EIS1600TMP = eis1600.markdown.convert_mARkdown_to_EIS1600TMP:main',
                  'disassemble_into_miu_files = eis1600.miu.disassemble_into_miu_files:main',
                  'fix_miu_annotation = eis1600.helper.fix_miu_annotation:main',
                  'insert_uids = eis1600.markdown.insert_uids:main',
                  'md_tags_to_bio = eis1600.helper.md_tags_to_bio:main',
                  'miu_random_revisions = eis1600.helper.miu_random_revisions:main',
                  'miu_stats = eis1600.stats.miu_stats:main',
                  'onomastic_annotation = eis1600.onomastics.annotation:main',
                  'toponym_annotation = eis1600.toponyms.annotation:main',
                  'reassemble_from_miu_files = eis1600.miu.reassemble_from_miu_files:main',
                  'update_uids = eis1600.markdown.update_uids:main',
                  'xx_update_uids_old_process = eis1600.markdown.update_uids_old_process:main',
                  'yml_to_json = eis1600.helper.yml_to_json:main'
          ],
      },
      python_requires='>=3.7, <3.9',
      install_requires=[
              'openiti',
              'pandas',
              'numpy',
              'tqdm',
              'p_tqdm',
              'importlib_resources',
              'jsonpickle'
      ],
      extras_require={'NER': ['camel-tools']},
      classifiers=['Programming Language :: Python :: 3',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: OS Independent',
                   'Development Status :: 1 - Planning',
                   'Intended Audience :: Science/Research']
      )



from sys import argv, exit
from os.path import isfile, splitext
from argparse import ArgumentParser, Action, RawDescriptionHelpFormatter
from functools import partial
from pathlib import Path

from p_tqdm import p_uimap
from tqdm import tqdm

from eis1600.helper.repo import get_files_from_eis1600_dir, read_files_from_readme, MIU_REPO
from eis1600.miu.methods import annotate_miu_file, get_mius


class CheckFileEndingAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.IDs' and fileext != '.EIS1600':
                parser.error('You need to input an IDs file or a single MIU file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


def main():
    arg_parser = ArgumentParser(
        prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
        description='''Script to NER annotate MIU file(s).
-----
Give an IDs file or a single MIU file as input
otherwise 
all files in the MIU directory are batch processed.
'''
        )
    arg_parser.add_argument('-v', '--verbose', action='store_true')
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument('-p', '--parallel', help='parallel processing', action='store_true')
    arg_parser.add_argument('-f', '--force', help='force re-annotation', action='store_true')
    arg_parser.add_argument(
        'input', type=str, nargs='?',
        help='IDs or MIU file to process',
        action=CheckFileEndingAction
        )
    args = arg_parser.parse_args()

    verbose = args.verbose
    force = args.force

    if args.input:
        infile = './' + args.input
        filepath, fileext = splitext(infile)
        if fileext == '.IDs':
            mius = get_mius(infile)[1:]  # First element is path to the OPENITI HEADER
            print(f'NER annotate MIUs of {infile}')
            if args.parallel:
                res = []
                res += p_uimap(partial(annotate_miu_file), mius)
            else:
                for miu in tqdm(mius):
                    try:
                        annotate_miu_file(miu)
                    except Exception as e:
                        print(miu, e)
        else:
            print(f'NER annotate {infile}')
            annotate_miu_file(infile, force_annotation=force)
    else:
        input_dir = MIU_REPO

        if not Path(input_dir).exists():
            print('Your working directory seems to be wrong, make sure it is set to the parent dir of '
                  '`EIS1600_MIUs/`.')
            exit()

        print(f'NER annotate MIU files')
        files_list = read_files_from_readme(input_dir, '# Texts disassembled into MIU files\n')
        infiles = get_files_from_eis1600_dir(input_dir, files_list, 'IDs')
        if not infiles:
            print('There are no IDs files to process')
            exit()

        for infile in infiles:
            if verbose:
                print(f'NER annotate MIUs of {infile}')

            mius = get_mius(infile)[1:]  # First element is path to the OPENITI HEADER
            if args.parallel:
                res = []
                res += p_uimap(partial(annotate_miu_file), mius)
            else:
                for miu in tqdm(mius):
                    try:
                        annotate_miu_file(miu)
                    except Exception as e:
                        print(miu, e)

    print('Done')
